{------------------------------------------------------------------------------}
{..............................................................................}
{------------------------------------------------------------------------------}

{

   Diese Unit wurde von Torsten Zuther im Rahmen eines Schlerpraktikums
   vom 5.5.2004 bis zum 19.5.2004 an der Universitt f. Informatik in Leipzig
   erstellt. Sie gehrt zum Projekt "Crypt".

   email: torsten.zuther@web.de

}

unit cryptunit;

interface

uses
 SysUtils, Dialogs, stdctrls;        // SysUtils fr Stringumwandlung
                           // Dialogs fr Messageboxen

type

  TRSARec = record         // enthlt alle notwendigen Werte fr eine RSA-Verschlsselung
    p,                     // 1. Primzahl
    q,                     // 2. Primzahl
    e,                     // Verschlsselungsexponent e (encrypt) -> teilerfremd zu (p-1)*(q-1)
    d,                     // modulare Inverse zu e (d - decrypt) oder Entschlsselungsexponent
    p1q1  : Integer;       // (p-1)*(q-1)
   end;


  TBlowArr = array[1..3] of Char;           // ein Datensatz fr den Aufblhen-Algorithmus



{ ---- folgende Functions/Types  werden nur vom DES-Algorithmus verwendet ---- }
{ ----          ************************************************               }

  TPermutationTab = array[1..64] of Byte;   // Permutationstabelle(n) fr den DES-Algorithmus

  SBox            = array[0..3,0..15]  of Byte;   // es gibt 8 S-Boxen beim DES-Algorithmus

  TDES_keys       = array[1..16] of string;



const

  IPTab    : TPermutationTab = (58,50,42,34,26,18,10,02,60,52,44,36,28,20,12,04,    // 64 bit
                                62,54,46,38,30,22,14,06,64,56,48,40,32,24,16,08,
                                57,49,41,33,25,17,09,01,59,51,43,35,27,19,11,03,
                                61,53,45,37,29,21,13,05,63,55,47,39,31,23,15,07);
  ExpTab   : TPermutationTab = (32,01,02,03,04,05,04,05,06,07,08,09,08,09,10,11,    // 48 bit
                                12,13,12,13,14,15,16,17,16,17,18,19,20,21,20,21,
                                22,23,24,25,24,25,26,27,28,29,28,29,30,31,32,01,
                                00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00);
  PC1Tab   : TPermutationTab = (57,49,41,33,25,17,09,01,58,50,42,34,26,18,10,02,    // 56 bit
                                59,51,43,35,27,19,11,03,60,52,44,36,63,55,47,39,
                                31,23,15,07,62,54,46,38,30,22,14,06,61,53,45,37,
                                29,21,13,05,28,20,12,04,00,00,00,00,00,00,00,00);
  PC2Tab   : TPermutationTab = (14,17,11,24,01,05,03,28,15,06,21,10,23,19,12,04,    // 48 bit
                                26,08,16,07,27,20,13,02,41,52,31,37,47,55,30,40,
                                51,45,33,48,44,49,39,56,34,53,46,42,50,36,29,32,
                                00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00);
  PTab     : TPermutationTab = (16,07,20,21,29,12,28,17,01,15,23,26,05,18,31,10,    // 32 bit
                                02,08,24,14,32,27,03,09,19,13,30,06,22,11,04,25,
                                00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,
                                00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00);
  PITab    : TPermutationTab = (40,08,48,16,56,24,64,32,39,07,47,15,55,23,63,31,    // 64 bit
                                38,06,46,14,54,22,62,30,37,05,45,13,53,21,61,29,
                                36,04,44,12,52,20,60,28,35,03,43,11,51,19,59,27,
                                34,02,42,10,50,18,58,26,33,01,41,09,49,17,57,25);

  S1       : SBox            = ((14,04,13,01,02,15,11,08,03,10,06,12,05,09,00,07),
                                (00,15,07,04,14,02,13,01,10,06,12,11,09,05,03,08),
                                (04,01,14,08,13,06,02,11,15,12,09,07,03,10,05,00),
                                (15,12,08,02,04,09,01,07,05,11,03,14,10,00,06,13));
  S2       : SBox            = ((15,01,08,14,06,11,03,04,09,07,02,13,12,00,05,10),
                                (03,13,04,07,15,02,08,14,12,00,01,10,06,09,11,05),
                                (00,14,07,11,10,04,13,01,05,08,12,06,09,03,02,15),
                                (13,08,10,01,03,15,04,02,11,06,07,12,00,05,14,09));
  S3       : SBox            = ((10,00,09,14,06,03,15,05,01,13,12,07,11,04,02,08),
                                (13,07,00,09,03,04,06,10,02,08,05,14,12,11,15,01),
                                (13,06,04,09,08,15,03,00,11,01,02,12,05,10,14,07),
                                (01,10,13,00,06,09,08,07,04,15,14,03,11,05,02,12));
  S4       : SBox            = ((07,13,14,03,00,06,09,10,01,02,08,05,11,12,04,15),
                                (13,08,11,05,06,15,00,03,04,07,02,12,01,10,14,09),
                                (10,06,09,00,12,11,07,13,15,01,03,14,05,02,08,04),
                                (03,15,00,06,10,01,13,08,09,04,05,11,12,07,02,14));
  S5       : SBox            = ((02,12,04,01,07,10,11,06,08,05,03,15,13,00,14,09),
                                (14,11,02,12,04,07,13,01,05,00,15,10,03,09,08,06),
                                (04,02,01,11,10,13,07,08,15,09,12,05,06,03,00,14),
                                (11,08,12,07,01,14,02,13,06,15,00,09,10,04,05,03));
  S6       : SBox            = ((12,01,10,15,09,02,06,08,00,13,03,04,14,07,05,11),
                                (10,15,04,02,07,12,09,05,06,01,13,14,00,11,03,08),
                                (09,14,15,05,02,08,12,03,07,00,04,10,01,13,11,06),
                                (04,03,02,12,09,05,15,10,11,14,01,07,06,00,08,13));
  S7       : SBox            = ((04,11,02,14,15,00,08,13,03,12,09,07,05,10,06,01),
                                (13,00,11,07,04,09,01,10,14,03,05,12,02,15,08,06),
                                (01,04,11,13,12,03,07,14,10,15,06,08,00,05,09,02),
                                (06,11,13,08,01,04,10,07,09,05,00,15,14,02,03,12));
  S8       : SBox            = ((13,02,08,04,06,15,11,01,10,09,03,14,05,00,12,07),
                                (01,15,13,08,10,03,07,04,12,05,06,11,00,14,09,02),
                                (07,11,04,01,09,12,14,02,00,06,10,13,15,03,05,08),
                                (02,01,14,07,04,10,08,13,15,12,09,00,03,05,06,11));

  ShlTab : array[1..16] of Integer = (1,1,2,2,2,2,2,2,1,2,2,2,2,2,2,1);   // Bitschiebetabelle (links )
  ShrTab : array[1..16] of Integer = (0,1,2,2,2,2,2,2,1,2,2,2,2,2,2,1);   // Bitschiebetabelle (rechts)


  function  Int64ToBin(i  : Int64)  : string;   // s. bytetobin
  function  BinToInt64(bin: string) : Int64;    // s. bintobyte


  function  ShiftL(count: Integer; s: string): string;     // verschiebt einen String nach links
  function  ShiftR(count: Integer; s: string): string;     // verschiebt einen String nach rechts

  function  _Xor(s,s2: string): string;                    // verknpft 2 binre strings mit XOR

  function  GenDESKeys(Encrypt: Boolean; key: string): TDES_keys;

  function  DES_Enc_datablock(data,key: string; flag: Boolean): string;     // DES-Algorithmus (ENCRYPT); flag immer true
  function  DES_Dec_datablock(data,key: string)               : string;     // DES_Algorithmus (DECRYPT)

  procedure DES_ENC(readfile,writefile: string; key: string; flag: Boolean);
  procedure DES_DEC(readfile,writefile: string; key: string);

  { ----          ************************************************          ---- }

  function  BoolAdd(b1,b2: Char)   : Char;     // addiert zwei Bits ('0' oder '1') binr
  function  BinToByte(bin: string): Byte;      // wandelt einen binren String (z.b. 01101101) in eine Dezimalzahl um
  function  ByteToBin(byt: Byte)   : String;   // wandelt eine Dezimalzahl in einen binren string

  procedure TauschChiffreEnc(s,t: Integer; readfile,writefile: string);         //  } \
  procedure TauschChiffreDec(s,t: Integer; readfile,writefile: string);         //  } /--> TAUSCH-CHIFFRE (kapselt Csar- und Multipli-
                                                                                //                         kationsverfahren ineinander)


  procedure CaesarEnc(s: Integer; readfile,writefile: string);                  // einfacher Csar-Algorithmus (Verschlsselung)
  procedure CaesarDec(s: Integer; readfile,writefile: string);                  // und wieder entschlsseln...

  procedure Multipl_enc(t: Integer; readfile,writefile: string);                // einfache Multiplikationsverschlsselung
  procedure Multipl_dec(t: Integer; readfile,writefile: string);                // einfache Multiplikationsentschlsselung
  function  Inv(m,t: integer): integer;                                         // Ermittelt die modulare Inverse von t (euklidischer Alg.)


  function  GenRandomKey(n,count: Integer; init: string; mark: Integer)  : string;
      // generiert mit der Lnge n (mind.3 hchstens 8) und dem Initialisierungsbinrstring init und
      // des markierten Bits ein lineares Schieberegister, was ein binren Schlssel herstellt


  function  CheckNullPosition(n: Integer; init: string; mark: Integer)   : Boolean;
      // prft ein Schieberegister auf eine Nullposition

  procedure BoolAdd_enc(n,mark: Integer; init,readfile,writefile: string);
      // verschlsselt durch Boolesche Addition (Schlssel wird mithilfe von GenRandomKey(n,count,init,mark) erzeugt.
      // ist gleichzeitig auch Entschlsselungsfunktion

  procedure Vigenere_enc(passwort, readfile,writefile: string);                 // verschlsselt mit Passwort nach dem
  procedure Vigenere_dec(passwort, readfile,writefile: string);                 // Csar-Verfahren

  procedure Xor_enc(x: Integer; readfile,writefile: string);                    // einfache Xor-Verschlsselung
  procedure Xor_dec(x: Integer; readfile,writefile: string);                    // Xor-Entschlsselung

  function  GenRandomRSARec: TRSARec;                                           // generiert Zufallsschlssel fr den RSA-Modus
  function  IsPrim(p: Integer): Boolean;                                        // Prft die Zahl p ob diese eine Primzahl ist

  procedure RSA_enc(e,n: Integer; readfile,writefile: string);                  // verschlsselt eine Datei mit dem RSA-Modus
  procedure RSA_dec(d,n: Integer; readfile, writefile: string);                 // entschlsselt eine Datei mit dem RSA-Modus

  procedure Blow_enc(readfile,writefile: string);                               // verschlsselt durch Aufblhen
  procedure Blow_dec(readfile,writefile: string);                               // entschlsselt es wieder


implementation

function Int64ToBin(i: Int64): string;
var
  j : Integer;
begin
  Result := '';
  for j := 63 downto 0 do Result := Result + IntToStr((i shr j) and 1);
end;


function  BinToInt64(bin: string): Int64;
var
 b: Int64;                    // Zwischenspeicherung
 i: Integer;                  // Laufvariable
begin
 b := 0;                                         // b initialisieren
 for i := 1 to 64  do                            // fr jedes der 64 Bits...
  if bin[i] = '1' then b := b or (1 shl (64-i)); // prfen ob es gesetzt ist.
                                                 // wenn ja dann wird es auch in b gesetzt
 Result := b;                                    // b zurckgeben
end;


function ShiftL(count: Integer; s: string): string;
var
 i,j   : Integer;
 res   : string;
 _1st  : Char;
begin
 res := s;
 for j := 1 to count do
  begin
   _1st := res[1];
   for i := 1 to Length(s)-1 do
    begin
     res[i] := res[i+1];
    end;
   res[Length(s)] := _1st;
  end;
 Result := res;
end;


function ShiftR(count: Integer; s: string): string;
var
 i,j   : Integer;
 res   : string;
 last  : Char;
begin
 res := s;
 for j := 1 to count do
  begin
   last := res[Length(res)];
   for i := Length(s) downto 1 do
    begin
     res[i] := res[i-1];
    end;
   res[1] := last;
  end;
 Result := res;
end;


function _Xor(s,s2: string): string;
var
 i   : Integer;
 res : string;
begin
 res := '';
 if Length(s) = Length(s2) then
  for i := 1 to Length(s) do
   if s[i] <> s2[i] then res := res + '1' else res := res + '0';
 Result := res;
end;

function GenDESKeys(Encrypt: Boolean; key: string): TDES_keys;
var
 i,j   : Integer;
 C,D   : string;
 keys  : TDES_keys;
 push  : string;
 key_s : string;
begin
 C     := Copy(key,01,28);
 D     := Copy(key,29,28);
 push  := key;
 key_s := '';

 for i := 1 to 56 do
  key_s := key_s + push[PC1Tab[i]];

 for i := 1 to 16 do
  begin
   if encrypt then
    begin
     C := ShiftL(shltab[i],C);              // Linksverschiebung
     D := ShiftL(shltab[i],D);
    end
   else
    begin
     C := ShiftR(shrtab[i],C);              // Rechtsverschiebung
     D := ShiftR(shrtab[i],D);
    end;
   push := C + D;                             // wieder Zusammenfgen
   keys[i] := '';
   for j := 1 to 48 do                        // Schlsselauswahlverfahren (PC2-Permutation)
    begin
     keys[i] := keys[i] + push[PC2Tab[j]];
    end;
  end;
 Result := keys;
end;


function  DES_Enc_datablock(data,key: string; flag: Boolean): string;  // verschlsselt eine 64bit Zahl mit einem
                                                        // 64bit Schlssel nach dem DES
var
 i            : Integer;  // Laufvariable
 data_string  : string;   // der 64bit-Datenblock als binrer String
 L,R,R_alt    : string;   // L und R-Register
 push,push2   : string;   // (Zwischenspeicher) fr String-Ops
 round        : Integer;  // Verschlsselungsrunde (1-16)
 sb1,sb2,sb3,
 sb4,sb5,sb6,
 sb7,sb8      : string;
 keys         : TDES_keys;
begin
 round := 1;
 data_string := data;
 push := data_string;
 data_string := '';
 for i := 1 to 64 do                         // Initiale Permutation
  begin
   data_string := data_string + push[IPTab[i]];
  end;
 L := Copy(data_string,1,32);                // Zerteilung in L- und R-Register
 R := Copy(data_string,33,32);
 R_alt := R;
 keys := GenDESKeys(flag,key);

// -----------------------------------------------------------------------------

while round <= 16 do

BEGIN

 push := R;
 R    := '';
 for i := 1 to 48 do                        // Expansionspermutation von R
 begin
  R := R + push[ExpTab[i]];
 end;

 R := _Xor(R, keys[round]);                  // R mit Teilschlssel verknpfen

// R durch die S-Boxen schicken...

 sb1 := Copy(R, 1, 6);
 sb2 := Copy(R, 7, 6);
 sb3 := Copy(R, 13, 6);
 sb4 := Copy(R, 19, 6);
 sb5 := Copy(R, 25, 6);
 sb6 := Copy(R, 31, 6);
 sb7 := Copy(R, 37, 6);
 sb8 := Copy(R, 43, 6);

 push  := '000000' + sb1[1] + sb1[6];
 push2 := '0000' + Copy(sb1,2,4);
 sb1 := Copy(ByteToBin(S1[BinToByte(push),BinToByte(push2)]),5,4);

 push  := '000000' + sb2[1] + sb2[6];
 push2 := '0000' + Copy(sb2,2,4);
 sb2 := Copy(ByteToBin(S2[BinToByte(push),BinToByte(push2)]),5,4);

 push  := '000000' + sb3[1] + sb3[6];
 push2 := '0000' + Copy(sb3,2,4);
 sb3 := Copy(ByteToBin(S3[BinToByte(push),BinToByte(push2)]),5,4);

 push  := '000000' + sb4[1] + sb4[6];
 push2 := '0000' + Copy(sb4,2,4);
 sb4 := Copy(ByteToBin(S4[BinToByte(push),BinToByte(push2)]),5,4);

 push  := '000000' + sb5[1] + sb5[6];
 push2 := '0000' + Copy(sb5,2,4);
 sb5 := Copy(ByteToBin(S5[BinToByte(push),BinToByte(push2)]),5,4);

 push  := '000000' + sb6[1] + sb6[6];
 push2 := '0000' + Copy(sb6,2,4);
 sb6 := Copy(ByteToBin(S6[BinToByte(push),BinToByte(push2)]),5,4);

 push  := '000000' + sb7[1] + sb7[6];
 push2 := '0000' + Copy(sb7,2,4);
 sb7 := Copy(ByteToBin(S7[BinToByte(push),BinToByte(push2)]),5,4);

 push  := '000000' + sb8[1] + sb8[6];
 push2 := '0000' + Copy(sb8,2,4);
 sb8 := Copy(ByteToBin(S8[BinToByte(push),BinToByte(push2)]),5,4);

 R := sb1+sb2+sb3+sb4+sb5+sb6+sb7+sb8;

 push := R;
 R    := '';

 for i := 1 to 32 do                        // Permutation von R
 begin
  R := R + push[PTab[i]];
 end;

 R := _Xor(R,L);

 L := R_alt;
 R_alt := R;

 Inc(round);
END;

// -----------------------------------------------------------------------------
 data_string := '';
 push := R+L;
 for i := 1 to 64 do                        // Permutation von data
 begin
  data_string := data_string + push[PITab[i]];
 end;

 Result := data_string;
end;


function  DES_Dec_datablock(data,key: string): string;  // verschlsselt eine 64bit Zahl mit einem
                                                        // 64bit Schlssel nach dem DES
begin
 Result := DES_Enc_datablock(data,key,false);
end;


procedure DES_ENC(readfile,writefile,key: string; flag: Boolean);
var
 i,j,k     : Integer;
 f,f2      : File of Byte;
 b         : Byte;
 data,s,s2 : string;
begin
 AssignFile(f, readfile);
 AssignFile(f2, writefile);
 Reset(f);
 Rewrite(f2);
 i := 0;
 data := '';
 while i < FileSize(f)-(FileSize(f) mod 8) do
  begin
   data := '';
   for j := i to i + 7 do
    begin
     Seek(f,j);
     Read(f,b);
     s  := '';
     s2 := '';
     s := s + ByteToBin(b);
     for k := 1 to 8 - Length(s) do
      s2 := s2 + '0';
     s := s2 + s;
     data := data + s;
    end;
   data := DES_enc_datablock(data,key,flag);
   for k := 0 to 7 do
   begin
    b := BinToByte(Copy(data,k*8+1,8));
    Write(f2, b);
   end;
   Inc(i,8);
  end;
 for i := FileSize(f)-(FileSize(f) mod 8) to FileSize(f)-1 do
  begin
   Seek(f,i);
   Read(f,b);
   Write(f2,b);
  end;
 CloseFile(f);
 CloseFile(f2);
end;


procedure DES_DEC(readfile,writefile: string; key: string);
begin
end;


procedure CaesarEnc(s: Integer; readfile,writefile: string);  // einfacher Csar-Algorithmus (Verschlsselung)
var
 i    : Integer;      // eine Laufvariable
 f,f2 : File of Char; // die beiden Dateien
 ch   : Char;         // das Zeichen, was bearbeitet wird
begin
 AssignFile(f, readfile);          // f mit Lesedatei verknpfen
 Reset(f);                         // f muss vorhanden sein
 AssignFile(f2, writefile);        // f2 mit Schreibdatei verknpfen
 Rewrite(f2);                      // f2 in jedem Fall erzeugen (auch berschreiben)
 for i := 0 to FileSize(f)-1 do    // fr jedes Byte oder ASCII-Zeichen
  begin
   Seek(f,i);                      // springe zu aktuellen Position
   Read(f,ch);                     // lies das Zeichen ein
   ch := Chr((Ord(ch)+s) mod 256); // ch um s verschieben
   Write(f2,ch);                   // verschlsseltes ch in Ausgabedatei schreiben
  end;
  CloseFile(f);                    // Dateien schlieen
  CloseFile(f2);
end;


procedure CaesarDec(s: Integer; readfile,writefile: string); // einfacher Csar-Algorithmus (Verschlsselung)
begin
 CaesarEnc(-s,readfile,writefile);  // ist das selbe wie die Csarverschlsselung, blo dass um -s Einheiten verschoben wird
end;


procedure Multipl_enc(t: Integer; readfile,writefile: string);   // einfache Multiplikationsverschlsselung
var
 i    : Integer;       // eine Laufvariable
 ch   : Char;          // das Zeichen, was bearbeitet wird
 f,f2 : File of Char;  // die beiden Dateien
begin
 AssignFile(f,readfile);                 // f mit Lesedatei verknpfen
 AssignFile(f2,writefile);               // f2 mit Schreibdatei verknpfen
 Reset(f);                               // f muss vorhanden sein
 Rewrite(f2);                            // f2 in jedem Fall erzeugen (auch berschreiben)
 if ((Inv(256,t)*t-1) mod 256) = 0 then  // wenn t ein legitimer multiplikativer Schlssel ist...
  begin
   for i := 0 to FileSize(f)-1 do        // fr jedes Byte oder ASCII-Zeichen
    begin
     Seek(f,i);                          // springe zu aktuellen Position
     Read(f,ch);                         // lies das Zeichen ein
     ch := Chr((Ord(ch)*t) mod 256);     // ch mit t multiplizieren (im Modul 256(ASCII-Zeichenanzahl) versteht sich)
     Write(f2,ch);                       // verschlsseltes ch in Ausgabedatei schreiben
    end;
  end
 else
  ShowMessage('Fehlerhafter Multiplikator!');
 CloseFile(f);                           // Dateien schlieen
 CloseFile(f2);
end;


procedure Multipl_dec(t: Integer; readfile,writefile: string);  // einfache Multiplikationsentschlsselung
begin
 Multipl_enc(Inv(256,t),readfile,writefile); // mit der modularen Inverse von t verschlsseln = mit t entschlsseln
end;


procedure TauschChiffreEnc(s,t: Integer; readfile,writefile: string);  //kapselt Csar- und Multiplikationsverfahren
var
 ch   : Char;          // das Zeichen, was bearbeitet wird
 i    : Integer;       // eine Laufvariable
 f,f2 : File of Char;  // die beiden Dateien
begin
 AssignFile(f,readfile);                 // f mit Lesedatei verknpfen
 AssignFile(f2,writefile);               // f2 mit Schreibdatei verknpfen
 Reset(f);                               // f muss vorhanden sein
 Rewrite(f2);                            // f2 in jedem Fall erzeugen (auch berschreiben)
 if ((Inv(256,t)*t-1) mod 256) = 0 then  // wenn t ein legitimer multiplikativer Schlssel ist...
  begin
   for i := 0 to FileSize(f)-1 do        // fr jedes Byte oder ASCII-Zeichen
    begin
     Seek(f, i);                         // springe zu aktuellen Position
     Read(f, ch);                        // lies das Zeichen ein
     ch := Chr(((Ord(ch)+s)*t) mod 256); // ch um s verschieben und mit t multiplizieren
     Write(f2,ch);                       // verschlsseltes ch in Ausgabedatei schreiben
    end;
  end
 else
  ShowMessage('Fehlerhafter Multiplikator!');
 CloseFile(f);                           // Dateien schlieen
 CloseFile(f2);
end;


procedure TauschChiffreDec(s,t: Integer; readfile,writefile: string);
var
 ch   : Char;           // das Zeichen, was bearbeitet wird
 i    : Integer;        // eine Laufvariable
 f,f2 : File of Char;   // die beiden Dateien
begin
 AssignFile(f,readfile);                 // f mit Lesedatei verknpfen
 AssignFile(f2,writefile);               // f2 mit Schreibdatei verknpfen
 Reset(f);                               // f muss vorhanden sein
 Rewrite(f2);                            // f2 in jedem Fall erzeugen (auch berschreiben)
 if ((Inv(256,t)*t-1) mod 256) = 0 then  // wenn t ein legitimer multiplikativer Schlssel ist...
  begin
   for i := 0 to FileSize(f)-1 do        // fr jedes Byte oder ASCII-Zeichen
    begin
     Seek(f, i);                         // springe zu aktuellen Position
     Read(f, ch);                        // lies das Zeichen ein
     ch := Chr(((Ord(ch)*Inv(256,t))-s) mod 256); // ch mit inv(t) multiplizueren und um -s verschieben
     Write(f2,ch);                       // entschlsseltes ch in Ausgabedatei schreiben
    end;
  end
 else
  ShowMessage('Fehlerhafter Multiplikator!');
 CloseFile(f);                           // Dateien schlieen
 CloseFile(f2);
end;


function Inv(m,t: integer): integer;    // Ermittelt die modulare Inverse von t (euklidischer Alg.)
var i: integer;                         // eine Laufvariable
begin
 for i := 1 to m-1 do                   // fr jedes 1 <= i <= m-1 ...
  begin
   if (t*i -1) mod m = 0 then           // wenn (t*i-1) ein Vielfaches von m ist...
    begin
     Result := i;                       // ...ist i die modulare Inverse von t im Modul m
     Exit;                              // bye bye
    end;
  end;
 Result := -1;                          // wenn es keine inv(m,t) gibt dann wird -1 zurckgegeben
end;


function BoolAdd(b1,b2: Char): Char;   // addiert zwei Bits ('0' oder '1') binr
begin
 Result := '0';
 if ((b1='1') and (b2='0')) or ((b1='0') and (b2='1')) then Result := '1';  // alles verstndlich !?
end;


function BinToByte(bin: string): Byte; // wandelt einen binren String (z.b. 01101101) in eine Dezimalzahl um
var
 b: Byte;                    // Zwischenspeicherung
 i: Integer;                 // Laufvariable
begin
 b := 0;                                        // b initialisieren
 for i := 1 to 8 do                             // fr jedes der 8 Bits...
  if bin[i] = '1' then b := b or (1 shl (8-i)); // prfen ob es gesetzt ist.
                                                // wenn ja dann wird es auch in b gesetzt
 Result := b;                                   // b zurckgeben
end;


function ByteToBin(byt: Byte): String;          // Gegenstck zu BinToByte
var
  i : Integer;
begin
  Result := '';
  for i := 7 downto 0 do Result := Result + IntToStr((byt shr i) and 1);
end;


function GenRandomKey(n,count: Integer; init: string; mark: Integer): string; // generiert binren Schlssel
var
 i,j          : Integer;                                               // Laufvariablen
 res, s, save : string;                                                // res= zu generierender string; s, save = Zwischenspeicher
begin
 res := '';                                                            // res initialisieren
 s := init;                                                            // init in s schreiben
 save := s;                                                            // save = s
 if (n>3) and (Length(init) = n) and ((mark < n) and (mark > 0)) then  // Bedingungen (verstndlich!?)
  begin
   for i := 1 to count do                                              // Periode (wie lang der Schlssel letzendlich wird)
    begin
     for j := 2 to n do
       s[j] := save[j-1];                                              // jedes Bit wird eins nach rechts geschoben
     s[1] := BoolAdd(save[Length(save)],save[mark]);                   // das erste Bit = letztes Bit + markiertes Bit
     res := res + save[Length(save)];                                  // res aktualisieren
     save := s;                                                        // save aktualisieren
    end;
  end;
 Result := res;                                                        // res zurckgeben
end;


function CheckNullPosition(n: Integer; init: string; mark: Integer): Boolean;  // checkt ein lin. Schieberegister auf Nullposititionen
var
 i,j          : Integer;   // Laufvars
 res, s, save : string;
begin
 Result := false;                                                      // Rckgabewert initialisieren
 s := init;                                                            // s bernimmt init
 save := s;                                                            // save = s
 if (n>3) and (Length(init) = n) and ((mark < n) and (mark > 0)) then  // Bedingungen fr das lin. Schieberegister
  begin
   for i := 1 to Trunc(exp(n*ln(2)))-1 do                              // fr alle Perioden (2^n) des Registers wird der Schlssel erzeugt
    begin
     for j := 2 to n do                                                // vom 2.Bit an wird dieses immer um eine Stelle nach rechts verschoben
       s[j] := save[j-1];
     s[1] := BoolAdd(save[Length(save)],save[mark]);                   // das erste, "freiwerdende Bit" entsteht aus Boolescher Addition vom letzten Bit und
                                                                       // dem markierten Bit (mark ist dafr der Index)
     res := res + save[Length(save)];                                  // das letzte Bit ist jeweils das Schlsselbit und wird an res drangehangen
     save := s;                                                        // save aktualisieren
     if StrToInt(save) = 0 then                                        // falls eine Nullposition entsteht...
      begin
       Result := true;                                                 // entsprechendes zurckgeben
       Exit;                                                           // und raus
      end;
    end;
  end;
end;


procedure BoolAdd_enc(n,mark: Integer; init,readfile,writefile: string);  // Verschlsselung durch boolesche Addition
var
 i,j,c: Integer;              // Lauf- und countvars
 b    : string;              // zu bearbeitendes Zeichen in Binrformat
 key  : string;               // generierter binrer Schlsselstring
 f,f2 : File of Char;         // Lese- und Schreibedatei
 ch   : Char;                 // zu bearbeitendes Zeichen
begin
 c   := -1;                                                             // c initialisieren
 if (n>3) and (Length(init) = n) and ((mark < n) and (mark > 0)) then   // Bedingungen fr das Register
 if not CheckNullPosition(n,init,mark) then                             // wenn keine Nullpositionen gefunden wurde, dann...
  begin
   key := GenRandomKey(n, Trunc(exp(n*ln(2))-1),init,mark);             // ...wird der Schlssel generiert
   MessageDlg('Folgender Binr-Schlssel wurde generiert:'+#13#10#13#10+key,mtInformation,[mbOK],0);  // und angezeigt
   AssignFile(f ,readfile);                                             // f mit Lesefile verknpfen
   Reset(f);                                                            // f muss vorhanden sein
   AssignFile(f2,writefile);                                            // f2 mit Schreibefile verknpfen
   Rewrite(f2);                                                         // f2 berschreiben
   for i := 0 to FileSize(f)-1 do                                       // fr jedes Zeichen...
    begin
     Seek(f,i);                                                         // zur aktuellen Position im Lesefile
     Read(f,ch);                                                        // Zeichen in ch einlesen
     b := ByteToBin(Ord(ch));                                           // ch in Binrcode umwandeln
     for j := 1 to 8 do                                                 // fr jedes der 8 Bits...
      begin
       Inc(c);                                                          // c erhhen (Counter)
       b[j] := BoolAdd(b[j],key[c mod Length(key)+1]);                  // Bit[index] von b mit dem entsprechen Bit des keys addieren
      end;
     ch := Chr(BinToByte(b));                                           // wieder zurck in ASCII-Format
     Write(f2,ch);                                                      // verschlsseltes ch in f2 schreiben
    end;
    CloseFile(f);                                                       // und die Dateien schlieen
    CloseFile(f2);
  end;
end;


procedure Vigenere_enc(passwort, readfile,writefile: string);   // Vigenere-Verschlsselung
var
 i   : Integer;              // Laufvar
 f,f2: File of Char;         // Dateien (lesen & schreiben)
 ch  : char;                 // zu bearbeitendes Zeichen
begin
 AssignFile(f,readfile);           //  f mit readfile verknpfen
 AssignFile(f2,writefile);         //  f2 mit writefile verknpfen
 Reset(f);                         //  f muss vorhanden sein
 Rewrite(f2);                      //  f2 wird evl. berschrieben
 for i := 0 to FileSize(f)-1 do    // fr jedes Byte
  begin
   Seek(f,i);                      // an entsprechenede Dateiposition
   Read(f, ch);                    // Zeichen einlesen...
   ch := Chr((Ord(passwort[i mod Length(passwort)+1]) + Ord(Ch)) mod 256);  // und um aktuelles Passwortchar verschieben
   Write(f2,ch);                   // verschlsseltes ch in f2 schreiben
  end;
 CloseFile(f);                     // Dateien schlieen
 CloseFile(f2);
end;


procedure Vigenere_dec(passwort, readfile,writefile: string);   // Vigenere-Entschlsselung
var
 s: string;     // Zwischenspeicher
 i: Integer;    // Laufvar
begin
 s := passwort;                        // passwort in s schreiben
 for i := 1 to Length(passwort) do     // jedes Zeichen in <passwort>...
  s[i] := Chr(256-Ord(passwort[i]));   // wird von 256 abgezogen...es entsteht eine Art "inverses Passwort"
 Vigenere_enc(s,readfile,writefile);   // dammit kann dann wiederum verschlsselt werden. damit wird entschlsselt!!
end;


procedure Xor_enc(x: Integer; readfile, writefile: string);    // Xor-Verschlsselung
var
 i    : Integer;         // Laufvar
 f,f2 : File of Byte;    // die Dateien wo gelesen & geschrieben wird
 b    : Byte;            // das aktuell zu bearbeitende Zeichen
begin
 AssignFile(f, readfile);     // f mit readfile verknpfen
 AssignFile(f2, writefile);   // f2 mit writefile verknpfen
 Reset(f);                    // f muss vorhanden sein
 Rewrite(f2);                 // f2 wird berschrieben
 for i := 0 to FileSize(f)-1 do    // jedes Byte...
  begin
   Seek(f,i);
   Read(f,b);                 // ... wird eingelesen
   b := b xor x;              // mit xor und dem geheimen Schlssel verknpft
   Write(f2,b);               // und in f2 geschrieben
  end;
 CloseFile(f);                // Dateien schlieen
 CloseFile(f2);
end;


procedure Xor_dec(x: Integer; readfile, writefile: string);    // Xor-Entschlsselung
begin
 Xor_enc(x,readfile,writefile);      // ist das selbe wie entschlsseln
end;


function IsPrim(p: Integer): Boolean;    // prft eine Zahl ob sie prim ist
var i: Integer;                          // Laufvar
begin
 Result := false;                        // erst mal keine Primzahl :(
 if p > 1 then                           // wenn p > 1 ist ...
  begin
   for i := 2 to Trunc(sqrt(p)) do       // werden alle Zahlen bis zur Quadratwurzel von p geprft...
    begin
     if p mod i = 0 then Exit;           // ob sie Teiler von p sind. wenn ja dann ist das keine Primzahl...
    end;
  end;
 Result := true;                         // ...wenn nein dann schon
end;


function GenRandomRSARec: TRSARec;      // generiert zufllige Schlssel fr das RSA-Verfahren
var
 r   : Integer;                         // Zwischenspeicher fr Zufallswerte
 res : TRSARec;                         // Result
label nochmal;                          // 3 Rcksprunglabels
label nochmal2;
label nochmal3;
begin
 Randomize;                      // Zufallsgenerator initialisieren
 r := Random(2000)+1;            // Zufallszahl im Bereich 1 < r 2000
nochmal:
 while (r < 2000) and (not IsPrim(r)) do Inc(r);   // solange erhhen bis r = prim ist
 if IsPrim(r) then
  begin
   res.p := r;                                     // wenn r = prim dann in p speichern
nochmal2:
   r := Random(2000)+1;                            // zufallszahl
   while (r < 2000) and (not IsPrim(r)) and (r<>res.p) do Inc(r);    // solange erhhen bis r = prim ist
   if IsPrim(r) then
    begin
     res.q := r;                                   // wenn r = prim ist dann in q speichern
    end
   else
    goto nochmal2;                                 // ansonsten nochmal versuchen (q)
  end
 else
  goto nochmal;                                    // ansonsten nochmal versuche (p)

 res.p1q1 := (res.p-1)*(res.q-1);                  // brauchen wir fr die Berechnung von d!

nochmal3:
 r := Random(2000)+1;           //  jetzt muss eine teilerfremde zahl zu (p-1)*(q-1) gefunden werden
                                //  das ist der fall wenn eine modulare Inverse vorhanden ist
 while (r<2147483647) do
  begin
   if ((Inv(res.p1q1,r)*r-1) mod (res.p1q1)) = 0 then Break else Inc(r);
  end;
 if ((Inv(res.p1q1,r)*r-1) mod (res.p1q1)) = 0 then res.e := r else goto nochmal3; // wenn eine gefunden wurde...
 res.d := Inv(res.p1q1,res.e);       // ... dann ist d die modulare Inverse von e;
 Result := res;                      // res zurckgeben
end;


procedure RSA_enc(e,n: Integer; readfile,writefile: string);     // DIE RSA-VERSCHLSSELUNG!
var
 i,c,j : Integer;              // Laufvars
 f     : File of Byte;         // readfile
 f2    : File of Int64;        // writefile
 b     : Byte;                 // aktuell bearbeitetes Zeichen
 w     : Int64;                // verschlsseltes Zeichen (in Form einer Zahl). wird in f2 geschrieben
begin
 AssignFile(f, readfile);      // f mit readfile verknpfen
 Reset(f);                     // f muss vorhanden sein
 AssignFile(f2, writefile);    // f2 mit writefile verknpfen
 Rewrite(f2);                  // f2 wird berschrieben
 if e mod 2 = 1 then c := e - 1 else c := e;
     {
       Um b zu verschlsseln muss  " (b ^ e) mod n " gerechnet werden.
       Da e unter Umstnden viel zu gro fr den Integerbereich wre,
       muss die Rechnung " (b ^ e) mod n " in e div 2 (evl. + 1 wenn
       e ungerade ist) Schritte  unterteilt werden. c ist diese Anzahl an Schritten.
     }
 if e > 2 then                      // Bedingung

 for i := 0 to FileSize(f) -1 do    // jedes Zeichen aus der Datei, die gelesen wird...
  begin
   Seek(f,i);                       // Dateiposition
   Read(f,b);                       // ein Byte einlesen

   w := (b*b) mod n;                // w initialisieren

   for j := 2 to (c div 2) do       // b ^ e   mod n   _
    begin                           //                  \
     w := (w*b*b) mod n;            //                   \
    end;                            //                   _\|
                                    //                     `

   if c = e - 1 then w := (w * b) mod n;  // wenn e ungerade dann kommt noch mal w*b als Faktor dazu!

   Write(f2,w);                // w in f2 schreiben
  end;
  CloseFile(f);                // Dateien schlieen
  CloseFile(f2);
end;


procedure RSA_dec(d,n: Integer; readfile, writefile: string);
var
 i,j,c  : Integer;             // Laufvars
 f      : File of Int64;       // readfile
 f2     : File of Byte;        // writefile
 b      : Byte;                // verschlsseltes Zeichen (in Form einer Zahl). wird in f2 geschrieben
 i64    : Int64;               // Zwischenspeicher
 r      : Int64;               // -   "   -
begin
 AssignFile(f, readfile);          // f mit readfile verknpfen
 Reset(f);                         // f muss vorhanden sein
 AssignFile(f2, writefile);        // f2 mit writefile verknpfen
 Rewrite(f2);                      // f2 wird berschrieben

 if d mod 2 = 1 then c := d - 1 else c := d;
      {
       Um r zu entschlsseln muss  " (r ^ d) mod n " gerechnet werden.
       Da d unter Umstnden viel zu gro fr den Integerbereich wre,
       muss die Rechnung " (r ^ d) mod n " in d div 2 (evl. + 1 wenn
       d ungerade ist) Schritte  unterteilt werden. c ist diese Anzahl an Schritten.
     }
 if d > 2 then                     // Bedingung...
 for i := 0 to FileSize(f) -1 do   // jede Zahl aus der Datei, die gelesen wird...
  begin
   Seek(f,i);                      // Dateiposition
   Read(f,r);                      // Zahl aus f in r einlesen
   i64 := (r*r) mod n;             // i64 initialisieren
   for j := 2 to (c div 2) do      // r ^ d   mod n   _
    begin                          //                  \
     i64 := (i64*r*r) mod n;       //                   \
    end;                           //                   _\|
                                   //                     `

   if c = d - 1 then i64 := (i64 * r) mod n;  // wenn d ungerade dann kommt noch mal i64*r als Faktor dazu!
   i64 := (i64) mod 256;                      // i64 in ASCII-Zeichensatz zurckverwandeln
   b := (i64);                                // in Byte speichern
   Write(f2,b);                               // und in File of Byte (f2) schreiben
  end;
  CloseFile(f);                               // Dateien schlieen
  CloseFile(f2);
end;


procedure Blow_enc(readfile,writefile: string);      // Verschlsseln durch Aufblhen...
var
 i,b   : Integer;          // Laufvars
 f,f2  : File of Char;     // Input- & Outputfile
 ch    : Char;             // Var zum Einlesen eines Zeichens
 ba    : TBlowArr;         // "Aufblhungs-Array"
begin
 AssignFile(f,readfile);     // f mit readfile verknpfen
 AssignFile(f2,writefile);   // f2 mit writefile verknpfen
 Reset(f);                   // f muss vorhanden sein
 Rewrite(f2);                // f2 wird berschrieben
 Randomize;                  // Zufallszahlengenerator generieren
 i := -1;                    // i initialisieren
 while i < FileSize(f)-1 do    // fr jedes Zeichen der einzulesenden Datei...
  begin
   Inc(i);                     // Dateipositionszhler erhhen
   Seek(f,i);                  // Dateiposition setzen
   Read(f,ch);                 // Zeichen einlesen
   ba[3] := ch;                // im dritten Feld des Blow-Arrays speichern...
   if (i < FileSize(f)-2) or (FileSize(f) mod 2 = 0) then   // wenn Dateigre gerade ist oder das vorletzte Zeichen erreicht ist...
    begin
     Inc(i);               // ... wird das nchste Zeichen...
     Seek(f,i);
     Read(f,ch);
     ba[1] := ch;          // ... ins erste Feld des Blowarray geschrieben
    end
   else
    begin
     ba[1] := ' ';  // ansonsten mit Leerzeichen auffllen
    end;
   ba[2] := Chr(Random(255)+1);  // das jeweils zweite Zeichen ist ein Zufallszeichen
   for b := 1 to 3 do
    begin
     Write(f2,ba[b]);            // blowarray in f2 schreiben
    end;
  end;
 CloseFile(f);                  // Dateien schlieen
 CloseFile(f2);
end;


procedure Blow_dec(readfile,writefile: string);    // Filtert die Originalzeichen aus einer aufgeblhten Datei wieder raus...
var
 i    : Integer;           // Laufvar
 f    : File of TBlowArr;  // Lesefile
 f2   : File of Char;      // Schreibfile
 ba   : TBlowArr;          // eingelesenes BlowArr
begin
 AssignFile(f,readfile);         // f mit readfile verknpfen
 AssignFile(f2,writefile);       // f2 mit writefile verknpfen
 Reset(f);                       // f muss vorhanden sein
 Rewrite(f2);                    // f2 wird berschrieben
 for i := 0 to FileSize(f)-1 do  // fr jedes BlowArr
  begin
   Seek(f,i);                    // Dateizeiger setzen
   Read(f, ba);                  // Blowarray in ba einlesen
   Write(f2,ba[3]);              // nur das dritte...
   Write(f2,ba[1]);              // und das erste Zeichen in diese Reihenfolge in f2 schreiben
  end;
 CloseFile(f);                   // Dateien schlieen
 CloseFile(f2);
end;

end.

{------------------------------------------------------------------------------}
{..............................................................................}
{------------------------------------------------------------------------------}
